"""Scenario 1 for unit-testing of stop-idle-sessions

In Scenario 1, there is a single active user with one SSH session and one
non-idle VNC session. The SSH session is actively tunneled into the VNC
session. Separately, the root user is SSH'd into the system (for observation)
and there is an idle GDM session running.
"""
# Don't worry about how many lines are in this file. It's basically just a
# dataset (albeit with actual Python objects).
# pylint: disable=too-many-lines


import datetime
from ipaddress import ip_address, IPv4Address, IPv6Address
from textwrap import dedent
from typing import Any, List, Mapping, Optional, Set, Tuple, Union
from unittest.mock import MagicMock, Mock

import stop_idle_sessions.logind
import stop_idle_sessions.main
import stop_idle_sessions.ps
import stop_idle_sessions.ss
import stop_idle_sessions.tty

from . import test_logind, test_main, test_ps, test_ss


class Scenario1LogindTestCase(test_logind.LogindTestCase):
    """Scenario1 unit testing for the logind module

    See the docstring for the test_scenario1 module for an overall description
    of Scenario 1 (single active user w/ VNC).

    This was generated by the following shell pipeline:
        loginctl | tail -n +2 | head -n -2 | awk '{print $1}' |
        xargs -L1 loginctl show-session
    """

    def _mock_gio_results_spec(self) -> Mapping[str, Mapping[str, str]]:
        """Input data to populate the mock Gio object (logind API)"""
        return {
            "1267": {
                "Id": "1267",
                "User": "1002",
                "TTY": "pts/2",
                "Scope": "session-1267.scope",
                "Leader": "952165"
            },
            "1301": {
                "Id": "1301",
                "User": "0",
                "TTY": "pts/1",
                "Scope": "session-1301.scope",
                "Leader": "994974"
            },
            "1337": {
                "Id": "1337",
                "User": "1002",
                "TTY": "pts/0",
                "Scope": "session-1337.scope",
                "Leader": "1050298"
            },
            "c1": {
                "Id": "c1",
                "User": "42",
                "TTY": "tty1",
                "Scope": "session-c1.scope",
                "Leader": "5655"
            }
        }

    def _expected_logind_sessions(self) -> List[Mapping[str, Any]]:
        """Expected set of logind session attributes to be returned"""
        return [
            {
                "session_id": "1267",
                "uid": 1002,
                "tty": "pts/2",
                "leader": 952165,
                "scope": "session-1267.scope",
                "scope_path": "/user.slice/user-1002.slice/session-1267.scope"
            },
            {
                "session_id": "1301",
                "uid": 0,
                "tty": "pts/1",
                "leader": 994974,
                "scope": "session-1301.scope",
                "scope_path": "/user.slice/user-0.slice/session-1301.scope"
            },
            {
                "session_id": "1337",
                "uid": 1002,
                "tty": "pts/0",
                "leader": 1050298,
                "scope": "session-1337.scope",
                "scope_path": "/user.slice/user-1002.slice/session-1337.scope"
            },
            {
                "session_id": "c1",
                "uid": 42,
                "tty": "tty1",
                "leader": 5655,
                "scope": "session-c1.scope",
                "scope_path": "/user.slice/user-42.slice/session-c1.scope"
            }
        ]


class Scenario1CgroupPidsTestCase(test_ps.CgroupPidsTestCase):
    """Scenario1 unit testing for the ps module

    See the docstring for the test_scenario1 module for an overall description
    of Scenario 1 (single active user w/ VNC).

    This was generated by the following shell pipeline:
        cat /sys/fs/cgroup/systemd/user.slice/user-1002.slice/session-1267.scope/cgroup.procs |
        xargs -L1 ps -o 'pid,args' --no-headers -p

    In Scenario 1, we deliberately choose the Xvnc session which has a TON of
    child processes, just to make sure that no parsing errors occur with these
    unusual commandlines.
    """

    def _mock_process_specs(self) -> Mapping[int, str]:
        """Input data to populate the filesystem mock for the cgroup sysfs"""
        # Yes, there are some very long lines here!
        # pylint: disable=line-too-long
        return {
            772211: "/usr/libexec/gvfsd-metadata",
            952570: "/usr/bin/Xvnc :1 -auth /u/wk/auser/.Xauthority -desktop computer:1 (auser) -fp catalogue:/etc/X11/fontpath.d -geometry 1024x768 -pn -rfbauth /u/wk/auser/.vnc/passwd -rfbport 5901 -localhost",
            952581: "/bin/sh /u/wk/auser/.vnc/xstartup",
            952582: "/usr/libexec/gnome-session-binary",
            952591: "dbus-launch --sh-syntax --exit-with-session",
            952592: "/usr/bin/dbus-daemon --syslog --fork --print-pid 5 --print-address 7 --session",
            952644: "/usr/libexec/at-spi-bus-launcher",
            952649: "/usr/bin/dbus-daemon --config-file=/usr/share/defaults/at-spi2/accessibility.conf --nofork --print-address 3",
            952652: "/usr/libexec/at-spi2-registryd --use-gnome-session",
            952656: "/usr/libexec/gvfsd",
            952663: "/usr/libexec/gvfsd-fuse /run/user/1002/gvfs -f -o big_writes",
            952715: "/usr/bin/gnome-keyring-daemon --start --components=pkcs11",
            952727: "/usr/bin/gnome-shell",
            952766: "ibus-daemon --xim --panel disable",
            952768: "/usr/libexec/xdg-permission-store",
            952770: "/usr/libexec/gnome-shell-calendar-server",
            952774: "/usr/libexec/ibus-dconf",
            952775: "/usr/libexec/ibus-extension-gtk3",
            952777: "/usr/libexec/ibus-x11 --kill-daemon",
            952779: "/usr/libexec/ibus-portal",
            952805: "/usr/libexec/evolution-source-registry",
            952817: "/usr/libexec/goa-daemon",
            952819: "/usr/libexec/dconf-service",
            952837: "/usr/libexec/goa-identity-service",
            952841: "/usr/libexec/gvfs-udisks2-volume-monitor",
            952853: "/usr/libexec/gvfs-mtp-volume-monitor",
            952861: "/usr/libexec/gvfs-goa-volume-monitor",
            952868: "/usr/libexec/gvfs-gphoto2-volume-monitor",
            952875: "/usr/libexec/gvfs-afc-volume-monitor",
            952885: "/usr/libexec/gsd-power",
            952888: "/usr/libexec/gsd-print-notifications",
            952889: "/usr/libexec/gsd-rfkill",
            952890: "/usr/libexec/gsd-screensaver-proxy",
            952891: "/usr/libexec/gsd-sharing",
            952892: "/usr/libexec/gsd-smartcard",
            952893: "/usr/libexec/gsd-sound",
            952894: "/usr/libexec/gsd-xsettings",
            952895: "/usr/libexec/gsd-subman",
            952898: "/usr/libexec/gsd-wacom",
            952904: "/usr/libexec/gsd-account",
            952905: "/usr/libexec/gsd-clipboard",
            952907: "/usr/libexec/gsd-a11y-settings",
            952909: "/usr/libexec/gsd-datetime",
            952910: "/usr/libexec/evolution-calendar-factory",
            952911: "/usr/libexec/gsd-color",
            952913: "/usr/libexec/gsd-keyboard",
            952919: "/usr/libexec/gsd-housekeeping",
            952921: "/usr/libexec/gsd-mouse",
            952935: "/usr/libexec/gsd-media-keys",
            952941: "/usr/libexec/ibus-engine-simple",
            952959: "/usr/libexec/gsd-printer",
            953009: "/usr/libexec/evolution-calendar-factory-subprocess --factory all --bus-name org.gnome.evolution.dataserver.Subprocess.Backend.Calendarx952910x2 --own-path /org/gnome/evolution/dataserver/Subprocess/Backend/Calendar/952910/2",
            953028: "/usr/libexec/evolution-addressbook-factory",
            953050: "/usr/libexec/evolution-addressbook-factory-subprocess --factory all --bus-name org.gnome.evolution.dataserver.Subprocess.Backend.AddressBookx953028x2 --own-path /org/gnome/evolution/dataserver/Subprocess/Backend/AddressBook/953028/2",
            953201: "/usr/libexec/gsd-disk-utility-notify",
            953207: "/usr/bin/gnome-software --gapplication-service",
            953209: "/usr/libexec/evolution/evolution-alarm-notify",
            953210: "/usr/libexec/tracker-miner-fs",
            953212: "/usr/libexec/tracker-miner-apps",
            953217: "/usr/libexec/tracker-store"
        }

    def _expected_process_objects(self) -> List[stop_idle_sessions.ps.Process]:
        """Expected set of processes to be parsed out of the cgroup sysfs"""
        # Yes, there are some very long lines here!
        # pylint: disable=line-too-long
        return [
            stop_idle_sessions.ps.Process(
                    pid=772211,
                    cmdline="/usr/libexec/gvfsd-metadata",
                    environ={}
            ),
            stop_idle_sessions.ps.Process(
                    pid=952570,
                    cmdline="/usr/bin/Xvnc :1 -auth /u/wk/auser/.Xauthority -desktop computer:1 (auser) -fp catalogue:/etc/X11/fontpath.d -geometry 1024x768 -pn -rfbauth /u/wk/auser/.vnc/passwd -rfbport 5901 -localhost",
                    environ={}
            ),
            stop_idle_sessions.ps.Process(
                    pid=952581,
                    cmdline="/bin/sh /u/wk/auser/.vnc/xstartup",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952582,
                    cmdline="/usr/libexec/gnome-session-binary",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952591,
                    cmdline="dbus-launch --sh-syntax --exit-with-session",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952592,
                    cmdline="/usr/bin/dbus-daemon --syslog --fork --print-pid 5 --print-address 7 --session",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952644,
                    cmdline="/usr/libexec/at-spi-bus-launcher",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952649,
                    cmdline="/usr/bin/dbus-daemon --config-file=/usr/share/defaults/at-spi2/accessibility.conf --nofork --print-address 3",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952652,
                    cmdline="/usr/libexec/at-spi2-registryd --use-gnome-session",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952656,
                    cmdline="/usr/libexec/gvfsd",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952663,
                    cmdline="/usr/libexec/gvfsd-fuse /run/user/1002/gvfs -f -o big_writes",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952715,
                    cmdline="/usr/bin/gnome-keyring-daemon --start --components=pkcs11",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952727,
                    cmdline="/usr/bin/gnome-shell",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952766,
                    cmdline="ibus-daemon --xim --panel disable",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952768,
                    cmdline="/usr/libexec/xdg-permission-store",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952770,
                    cmdline="/usr/libexec/gnome-shell-calendar-server",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952774,
                    cmdline="/usr/libexec/ibus-dconf",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952775,
                    cmdline="/usr/libexec/ibus-extension-gtk3",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952777,
                    cmdline="/usr/libexec/ibus-x11 --kill-daemon",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952779,
                    cmdline="/usr/libexec/ibus-portal",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952805,
                    cmdline="/usr/libexec/evolution-source-registry",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952817,
                    cmdline="/usr/libexec/goa-daemon",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952819,
                    cmdline="/usr/libexec/dconf-service",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952837,
                    cmdline="/usr/libexec/goa-identity-service",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952841,
                    cmdline="/usr/libexec/gvfs-udisks2-volume-monitor",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952853,
                    cmdline="/usr/libexec/gvfs-mtp-volume-monitor",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952861,
                    cmdline="/usr/libexec/gvfs-goa-volume-monitor",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952868,
                    cmdline="/usr/libexec/gvfs-gphoto2-volume-monitor",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952875,
                    cmdline="/usr/libexec/gvfs-afc-volume-monitor",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952885,
                    cmdline="/usr/libexec/gsd-power",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952888,
                    cmdline="/usr/libexec/gsd-print-notifications",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952889,
                    cmdline="/usr/libexec/gsd-rfkill",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952890,
                    cmdline="/usr/libexec/gsd-screensaver-proxy",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952891,
                    cmdline="/usr/libexec/gsd-sharing",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952892,
                    cmdline="/usr/libexec/gsd-smartcard",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952893,
                    cmdline="/usr/libexec/gsd-sound",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952894,
                    cmdline="/usr/libexec/gsd-xsettings",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952895,
                    cmdline="/usr/libexec/gsd-subman",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952898,
                    cmdline="/usr/libexec/gsd-wacom",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952904,
                    cmdline="/usr/libexec/gsd-account",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952905,
                    cmdline="/usr/libexec/gsd-clipboard",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952907,
                    cmdline="/usr/libexec/gsd-a11y-settings",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952909,
                    cmdline="/usr/libexec/gsd-datetime",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952910,
                    cmdline="/usr/libexec/evolution-calendar-factory",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952911,
                    cmdline="/usr/libexec/gsd-color",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952913,
                    cmdline="/usr/libexec/gsd-keyboard",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952919,
                    cmdline="/usr/libexec/gsd-housekeeping",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952921,
                    cmdline="/usr/libexec/gsd-mouse",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952935,
                    cmdline="/usr/libexec/gsd-media-keys",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952941,
                    cmdline="/usr/libexec/ibus-engine-simple",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=952959,
                    cmdline="/usr/libexec/gsd-printer",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953009,
                    cmdline="/usr/libexec/evolution-calendar-factory-subprocess --factory all --bus-name org.gnome.evolution.dataserver.Subprocess.Backend.Calendarx952910x2 --own-path /org/gnome/evolution/dataserver/Subprocess/Backend/Calendar/952910/2",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953028,
                    cmdline="/usr/libexec/evolution-addressbook-factory",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953050,
                    cmdline="/usr/libexec/evolution-addressbook-factory-subprocess --factory all --bus-name org.gnome.evolution.dataserver.Subprocess.Backend.AddressBookx953028x2 --own-path /org/gnome/evolution/dataserver/Subprocess/Backend/AddressBook/953028/2",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953201,
                    cmdline="/usr/libexec/gsd-disk-utility-notify",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953207,
                    cmdline="/usr/bin/gnome-software --gapplication-service",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953209,
                    cmdline="/usr/libexec/evolution/evolution-alarm-notify",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953210,
                    cmdline="/usr/libexec/tracker-miner-fs",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953212,
                    cmdline="/usr/libexec/tracker-miner-apps",
                    environ={
                        'DISPLAY': ':1'
                    }
            ),
            stop_idle_sessions.ps.Process(
                    pid=953217,
                    cmdline="/usr/libexec/tracker-store",
                    environ={
                        'DISPLAY': ':1'
                    }
            )
        ]


class Scenario1LoopbackConnectionTestCase(test_ss.LoopbackConnectionTestCase):
    """Scenario1 unit testing for the ss module

    See the docstring for the test_scenario1 module for an overall description
    of Scenario 1 (single active user w/ VNC).

    This was generated by the following shell pipeline:
        /usr/sbin/ss --all --no-header --numeric --oneline --processes --tcp
    """

    def _mock_raw_ss_output(self) -> str:
        """Input data to populate the mocked network connections table"""
        return dedent("""\
            LISTEN    0      100         127.0.0.1:25           0.0.0.0:*     users:(("master",pid=5337,fd=14))                          
            LISTEN    0      5           127.0.0.1:5901         0.0.0.0:*     users:(("Xvnc",pid=952570,fd=6))                           
            LISTEN    0      128           0.0.0.0:111          0.0.0.0:*     users:(("rpcbind",pid=4410,fd=4),("systemd",pid=1,fd=42))  
            LISTEN    0      128           0.0.0.0:22           0.0.0.0:*     users:(("sshd",pid=4960,fd=3))                             
            LISTEN    0      2048        127.0.0.1:631          0.0.0.0:*     users:(("cupsd",pid=162159,fd=8))                          
            ESTAB     0      452        10.0.0.169:22        10.0.3.209:57343 users:(("sshd",pid=1256518,fd=4),("sshd",pid=1256491,fd=4))
            SYN-SENT  0      1          10.0.0.169:45198 151.101.193.91:443   users:(("gnome-software",pid=1259638,fd=22))               
            TIME-WAIT 0      0          10.0.0.169:41180     10.0.4.244:636                                                              
            TIME-WAIT 0      0          10.0.0.169:53546     10.0.2.100:3128                                                             
            ESTAB     0      0           127.0.0.1:5901       127.0.0.1:49688 users:(("Xvnc",pid=952570,fd=23))                          
            SYN-SENT  0      1          10.0.0.169:40676 151.101.129.91:443   users:(("gnome-shell",pid=1259171,fd=34))                  
            ESTAB     0      0          10.0.0.169:22         10.0.1.53:41516 users:(("sshd",pid=1259753,fd=4),("sshd",pid=1259733,fd=4))
            ESTAB     0      0           127.0.0.1:49688      127.0.0.1:5901  users:(("sshd",pid=1256518,fd=7))                          
            ESTAB     0      0          10.0.0.169:22         10.0.1.53:39700 users:(("sshd",pid=1050325,fd=4),("sshd",pid=1050298,fd=4))
            TIME-WAIT 0      0          10.0.0.169:41178     10.0.4.244:636                                                              
            ESTAB     0      0          10.0.0.169:725       10.0.1.202:2049                                                             
            TIME-WAIT 0      0          10.0.0.169:53532     10.0.2.100:3128                                                             
            ESTAB     0      0          10.0.0.169:22         10.0.1.53:54592 users:(("sshd",pid=995013,fd=4),("sshd",pid=994974,fd=4))  
            TIME-WAIT 0      0          10.0.0.169:41196     10.0.4.244:636                                                              
            TIME-WAIT 0      0          10.0.0.169:41176     10.0.4.244:636                                                              
            SYN-SENT  0      1          10.0.0.169:49948 151.101.193.91:443   users:(("gnome-shell",pid=6978,fd=32))                     
            ESTAB     0      612        10.0.0.169:22        10.0.3.209:57353 users:(("sshd",pid=1258236,fd=4),("sshd",pid=1258006,fd=4))
            TIME-WAIT 0      0          10.0.0.169:53550     10.0.2.100:3128                                                             
            ESTAB     0      0          10.0.0.169:861       10.0.1.203:2049                                                             
            LISTEN    0      100             [::1]:25              [::]:*     users:(("master",pid=5337,fd=15))                          
            LISTEN    0      5               [::1]:5901            [::]:*     users:(("Xvnc",pid=952570,fd=7))                           
            LISTEN    0      128              [::]:111             [::]:*     users:(("rpcbind",pid=4410,fd=6),("systemd",pid=1,fd=44))  
            LISTEN    0      2048            [::1]:631             [::]:*     users:(("cupsd",pid=162159,fd=7))                          
        """)

    def _expected_listening_ports(self) -> Set[int]:
        """Expected set of listening ports parsed out of the mock ss data"""
        return set([22, 25, 111, 631, 5901])

    def _expected_peer_pairs(self) -> Set[Tuple[Union[IPv4Address,
                                                      IPv6Address], int]]:
        """Expected set of remote peers parsed from the established conns"""
        return set([
                (ip_address('10.0.1.53'), 39700),
                (ip_address('10.0.1.53'), 41516),
                (ip_address('10.0.1.53'), 54592),
                (ip_address('10.0.1.202'), 2049),
                (ip_address('10.0.1.203'), 2049),
                (ip_address('10.0.3.209'), 57343),
                (ip_address('10.0.3.209'), 57353),
                (ip_address('127.0.0.1'), 5901),
                (ip_address('127.0.0.1'), 49688)
        ])

    def _expected_connections(self) -> List[stop_idle_sessions.ss.LoopbackConnection]:
        """Expected set of loopback connections identified from these conns"""
        return [
            stop_idle_sessions.ss.LoopbackConnection(
                    client=stop_idle_sessions.ss.Socket(
                        addr=ip_address('127.0.0.1'),
                        port=49688,
                        processes=[stop_idle_sessions.ps.Process(
                            pid=1256518,
                            cmdline="",
                            environ={}
                        )]
                    ),
                    server=stop_idle_sessions.ss.Socket(
                        addr=ip_address('127.0.0.1'),
                        port=5901,
                        processes=[stop_idle_sessions.ps.Process(
                            pid=952570,
                            cmdline="",
                            environ={}
                        )]
                    )
            )
        ]


class Scenario1MainLoopTestCase(test_main.MainLoopTestCase):
    """Scenario1 unit testing for the main module

    See the docstring for the test_scenario1 module for an overall description
    of Scenario 1 (single active user w/ VNC).

    See also the other Scenario1* TestCases for hints about how to generate
    the input data which was used to specify this test fixture.
    """

    def _mock_get_all_sessions(self) -> List[stop_idle_sessions.logind.Session]:
        """Input data to mock out the logind module and D-Bus"""
        return [
            Mock(spec_set=stop_idle_sessions.logind.Session,
                 session_id="1267",
                 session_type="tty",
                 uid=1002,
                 tty="pts/2",
                 leader=952165,
                 scope="session-1267.scope",
                 scope_path="/user.slice/user-1002.slice/session-1267.scope"),
            Mock(spec_set=stop_idle_sessions.logind.Session,
                 session_id="1301",
                 session_type="tty",
                 uid=0,
                 tty="pts/1",
                 leader=994974,
                 scope="session-1301.scope",
                 scope_path="/user.slice/user-0.slice/session-1301.scope"),
            Mock(spec_set=stop_idle_sessions.logind.Session,
                 session_id="1337",
                 session_type="tty",
                 uid=1002,
                 tty="pts/0",
                 leader=1050298,
                 scope="session-1337.scope",
                 scope_path="/user.slice/user-1002.slice/session-1337.scope"),
            Mock(spec_set=stop_idle_sessions.logind.Session,
                 session_id="c1",
                 session_type="wayland",
                 uid=42,
                 tty="tty1",
                 leader=5655,
                 scope="session-c1.scope",
                 scope_path="/user.slice/user-42.slice/session-c1.scope")
        ]

    def _mock_find_loopback_connections(self) -> List[stop_idle_sessions.ss.LoopbackConnection]:
        """Input data to mock out the ss utility and module"""
        return [
            stop_idle_sessions.ss.LoopbackConnection(
                client=stop_idle_sessions.ss.Socket(
                    addr=ip_address('127.0.0.1'),
                    port=49688,
                    processes=[stop_idle_sessions.ps.Process(
                        pid=1256518,
                        cmdline="",
                        environ={})
                    ]
                ),
                server=stop_idle_sessions.ss.Socket(
                    addr=ip_address('127.0.0.1'),
                    port=5901,
                    processes=[stop_idle_sessions.ps.Process(
                        pid=952570,
                        cmdline="",
                        environ={})
                    ]
                )
            )
        ]

    def _mock_processes_in_scope_path(self,
                                      scope_path: str) -> List[stop_idle_sessions.ps.Process]:
        """Input data to mock out the ps and cgroup interface module"""

        if scope_path == "/user.slice/user-1002.slice/session-1267.scope":
            pids = list(map(lambda pid: stop_idle_sessions.ps.Process(pid=pid,
                                                                      cmdline="",
                                                                      environ={}),
                            [772211, 952582, 952591, 952592, 952644, 952649,
                             952652, 952656, 952663, 952715, 952727, 952766,
                             952768, 952770, 952774, 952775, 952777, 952779,
                             952805, 952817, 952819, 952837, 952841, 952853,
                             952861, 952868, 952875, 952885, 952888, 952889,
                             952890, 952891, 952892, 952893, 952894, 952895,
                             952898, 952904, 952905, 952907, 952909, 952910,
                             952911, 952913, 952919, 952921, 952935, 952941,
                             952959, 953009, 953028, 953050, 953201, 953207,
                             953209, 953210, 953212, 953217]))

            pids.append(stop_idle_sessions.ps.Process(
                    pid=952570,
                    cmdline=("/usr/bin/Xvnc :1 -auth /u/wk/auser/.Xauthority "
                             "-desktop computer:1 (auser) -fp "
                             "catalogue:/etc/X11/fontpath.d -geometry "
                             "1024x768 -pn -rfbauth /u/wk/auser/.vnc/passwd "
                             "-rfbport 5901 -localhost"),
                    environ={}
            ))
            pids.append(stop_idle_sessions.ps.Process(
                    pid=952581,
                    cmdline="/bin/sh /u/wk/auser/.vnc/xstartup",
                    environ={'DISPLAY': ":1"}
            ))

            return pids

        if scope_path == "/user.slice/user-0.slice/session-1301.scope":
            return list(map(lambda pid: stop_idle_sessions.ps.Process(pid=pid,
                                                                      cmdline="",
                                                                      environ={}),
                            [1258996, 1259009, 1259010, 1259026, 1259028,
                             1259067, 1259083, 1259088, 1259093, 1259097,
                             1259105, 1259157, 1259171, 1259210, 1259212,
                             1259214, 1259218, 1259219, 1259221, 1259223,
                             1259235, 1259257, 1259268, 1259271, 1259280,
                             1259285, 1259290, 1259297, 1259305, 1259307,
                             1259308, 1259309, 1259310, 1259311, 1259312,
                             1259313, 1259314, 1259315, 1259321, 1259322,
                             1259325, 1259328, 1259329, 1259331, 1259337,
                             1259342, 1259343, 1259347, 1259349, 1259350,
                             1259366, 1259438, 1259452, 1259454, 1259473,
                             1259632, 1259638, 1259640]))

        if scope_path == "/user.slice/user-1002.slice/session-1337.scope":
            return list(map(lambda pid: stop_idle_sessions.ps.Process(pid=pid,
                                                                      cmdline="",
                                                                      environ={}),
                            [1050298, 1256518, 1256520]))

        if scope_path == "/user.slice/user-42.slice/session-c1.scope":
            return list(map(lambda pid: stop_idle_sessions.ps.Process(pid=pid,
                                                                      cmdline="",
                                                                      environ={}),
                            [5655, 5875, 5877, 6221, 6243, 6263, 6544, 6604,
                             6620, 6978, 9150, 9273, 9279, 9283, 9670, 10363,
                             10375, 10377, 10383, 10396, 10418, 10422, 10426,
                             10443, 10448, 10474, 10483, 10484, 10492, 10493,
                             10494, 10495, 10518]))

        raise KeyError(f'Unexpected scope argument: {scope_path}')

    def _mock_uid_to_username(self, uid: int) -> str:
        """Convert numeric UIDs to symbolic usernames"""
        return {0: 'root', 42: 'gdm', 1002: 'auser'}[uid]

    def _mock_retrieve_idle_time(self,
                                 display: str,
                                 xauthority: Optional[str]) -> Optional[datetime.timedelta]:
        """Mock information about an X11 DISPLAY's idle timeout"""
        assert xauthority is None or isinstance(xauthority, str)

        if display == ':0':
            return datetime.timedelta(seconds=47 * 60)
        if display == ':1':
            return datetime.timedelta(seconds=35)

        raise KeyError(f'Unexpected DISPLAY argument: {display}')

    def _mock_tty(self, name: str) -> stop_idle_sessions.tty.TTY:
        """Mock information about a tty/pty's timestamps"""

        if name == 'pts/2':
            atty = Mock(spec=stop_idle_sessions.tty.TTY,
                        full_name='/dev/pts/2',
                        atime=datetime.datetime(2024, 11, 22,
                                                6, 15, 0, 0,
                                                tzinfo=datetime.timezone.utc),
                        mtime=datetime.datetime(2024, 11, 22,
                                                6, 16, 0, 0,
                                                tzinfo=datetime.timezone.utc))
            atty.configure_mock(name=name)
            return atty

        if name == 'pts/1':
            atty = Mock(spec=stop_idle_sessions.tty.TTY,
                        name=name,
                        full_name='/dev/pts/1',
                        atime=datetime.datetime(2024, 11, 22,
                                                5, 46, 0, 0,
                                                tzinfo=datetime.timezone.utc),
                        mtime=datetime.datetime(2024, 11, 22,
                                                5, 46, 0, 0,
                                                tzinfo=datetime.timezone.utc))
            atty.configure_mock(name=name)
            return atty

        if name == 'pts/0':
            atty = Mock(spec=stop_idle_sessions.tty.TTY,
                        name=name,
                        full_name='/dev/pts/0',
                        atime=datetime.datetime(2024, 11, 22,
                                                6, 15, 0, 0,
                                                tzinfo=datetime.timezone.utc),
                        mtime=datetime.datetime(2024, 11, 22,
                                                6, 14, 0, 0,
                                                tzinfo=datetime.timezone.utc))
            atty.configure_mock(name=name)
            return atty

        if name == 'tty1':
            atty = Mock(spec=stop_idle_sessions.tty.TTY,
                        name=name,
                        full_name='/dev/tty1',
                        atime=datetime.datetime(2024, 11, 22,
                                                5, 30, 0, 0,
                                                tzinfo=datetime.timezone.utc),
                        mtime=datetime.datetime(2024, 11, 22,
                                                5, 30, 0, 0,
                                                tzinfo=datetime.timezone.utc))
            atty.configure_mock(name=name)
            return atty

        raise KeyError(f'Unexpected tty/pty name: {name}')

    def _now(self) -> datetime.datetime:
        """Return a fixed point in time to compare timedeltas against"""
        return datetime.datetime(2024, 11, 22,
                                 6, 17, 0, 0,
                                 tzinfo=datetime.timezone.utc)

    def _excluded_users(self) -> List[str]:
        """Supplement session assertion testing with a set of excluded users"""
        return []

    def _expected_sessions(self) -> List[stop_idle_sessions.main.Session]:
        """Register the expected set of fully-fleshed-out session objects"""

        logind_sessions = self._mock_get_all_sessions()

        session_1267 = stop_idle_sessions.main.Session(
                session=logind_sessions[0],
                tty=self._mock_tty('pts/2'),
                display=':1',
                display_idle=datetime.timedelta(seconds=35),
                username='auser',
                processes=[stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=772211,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952570,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952581,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952582,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952591,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952592,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952644,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952649,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952652,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952656,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952663,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952715,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952727,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952766,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952768,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952770,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952774,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952775,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952777,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952779,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952805,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952817,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952819,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952837,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952841,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952853,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952861,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952868,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952875,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952885,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952888,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952889,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952890,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952891,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952892,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952893,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952894,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952895,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952898,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952904,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952905,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952907,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952909,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952910,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952911,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952913,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952919,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952921,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952935,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952941,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=952959,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953009,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953028,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953050,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953201,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953207,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953209,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953210,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953212,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=953217,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[])]
        )

        session_1301 = stop_idle_sessions.main.Session(
                session=logind_sessions[1],
                tty=self._mock_tty('pts/1'),
                display=None,
                display_idle=None,
                username='root',
                processes=[stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1258996,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259009,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259010,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259026,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259028,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259067,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259083,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259088,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259093,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259097,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259105,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259157,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259171,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259210,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259212,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259214,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259218,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259219,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259221,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259223,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259235,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259257,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259268,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259271,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259280,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259285,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259290,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259297,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259305,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259307,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259308,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259309,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259310,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259311,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259312,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259313,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259314,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259315,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259321,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259322,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259325,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259328,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259329,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259331,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259337,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259342,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259343,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259347,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259349,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259350,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259366,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259438,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259452,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259454,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259473,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259632,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259638,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1259640,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[])]
        )

        session_1337 = stop_idle_sessions.main.Session(
                session=logind_sessions[2],
                tty=self._mock_tty('pts/0'),
                display=None,
                display_idle=None,
                username='auser',
                processes=[stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1050298,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1256518,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[
                                    stop_idle_sessions.ps.Process(pid=952570,
                                                                  cmdline="",
                                                                  environ={})
                                ],
                                tunneled_sessions=[session_1267]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=1256520,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[])]
        )

        session_c1 = stop_idle_sessions.main.Session(
                session=logind_sessions[3],
                tty=self._mock_tty('tty1'),
                display=':0',
                display_idle=datetime.timedelta(seconds=47 * 60),
                username='gdm',
                processes=[stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=5655,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=5875,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=5877,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6221,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6243,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6263,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6544,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6604,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6620,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=6978,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=9150,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=9273,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=9279,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=9283,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=9670,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10363,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10375,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10377,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10383,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10396,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10418,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10422,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10426,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10443,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10448,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10474,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10483,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10484,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10492,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10493,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10494,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10495,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[]),
                           stop_idle_sessions.main.SessionProcess(
                                process=stop_idle_sessions.ps.Process(pid=10518,
                                                                      cmdline="",
                                                                      environ={}),
                                tunneled_processes=[],
                                tunneled_sessions=[])]
        )

        return [session_1267, session_1301, session_1337, session_c1]

    def _expected_results(self) -> List[Tuple[bool, bool,
                                              Optional[datetime.timedelta]]]:
        """Specify the expected handling for each of the _expected_sessions

        The returned tuple for _expected_results is as follows:
          (skipped_exempt: bool, terminated: bool, idle_metric: timedelta)
        """

        return [
                # session ID = 1267
                # 35-seconde idle determined by DISPLAY=:1
                (False, False, datetime.timedelta(seconds=35)),

                # session ID = 1301
                # 31-minute idle (terminated) based on pty atime/mtime
                (False, True, datetime.timedelta(seconds=31 * 60)),

                # session ID = 1337
                # 1-minute idle determined by session 1267
                (False, False, datetime.timedelta(seconds=35)),

                # session ID = c1
                # 47-minute idle (NOT terminated) based on DISPLAY=:0
                (True, False, datetime.timedelta(seconds=47 * 60))
        ]



"""Common exception for indicating problems when parsing session info"""


class SessionParseError(Exception):
    """Common exception for indicating problems when parsing session info"""

    def __init__(self, message):
        self.message = message



"""Common logind testing logic shared across all scenarios"""


from functools import partial
import re
from typing import Any, List, Mapping
from unittest import TestCase, TestSuite
from unittest.mock import ANY, Mock, patch

from stop_idle_sessions.list_set import compare_list_sets
import stop_idle_sessions.logind


_SESSION_NODE_RE = r'^/org/freedesktop/login1/session/([^/]+)$'


class LogindTestCase(TestCase):
    """Unit testing for the logind module

    This TestCase is meant to be subclassed, NOT run directly. The load_tests
    function at the bottom of this module prevents it from being
    auto-discovered.
    """

    #
    # Subclasses need to override these methods
    #

    def _mock_gio_results_spec(self) -> Mapping[str, Mapping[str, str]]:
        """Subclasses should override this method"""
        raise NotImplementedError('_mock_gio')

    def _expected_logind_sessions(self) -> List[Mapping[str, Any]]:
        """Subclasses should override this method"""
        raise NotImplementedError('_expected_logind_sessions')

    #
    # Here are the actual test case methods -- these aren't usually overridden
    #

    def setUp(self):
        self._sessions = {}

        self._mocked_gio = self._mock_gio(self._mock_gio_results_spec())
        mock_gio_obj_patch = patch('stop_idle_sessions.logind.Gio',
                                   new=self._mocked_gio)
        mock_gio_obj_patch.start()
        self.addCleanup(mock_gio_obj_patch.stop)

    def test_logind_interface_parsed_objects(self):
        """Ensure that objects are appropriately parsed from the logind API"""

        sessions = list(stop_idle_sessions.logind.get_all_sessions())
        expected_logind_sessions = self._expected_logind_sessions()

        attrs = set()
        for expected_logind_session in expected_logind_sessions:
            attrs.update(expected_logind_session.keys())

        actual_logind_sessions: List[Mapping[str, Any]] = []
        for session in sessions:
            actual_logind_session: Mapping[str, Any] = {}
            for attr in attrs:
                actual_logind_session[attr] = getattr(session, attr)
            actual_logind_sessions.append(actual_logind_session)
        self._assert_fully_exercised()

        self.assertTrue(compare_list_sets(expected_logind_sessions,
                                            actual_logind_sessions))

    #
    # Internal methods used by test cases -- these should not be overridden
    #

    def _mock_gio(self, results_spec: Mapping[str, Mapping[str, Any]]) -> Mock:
        """Mock the Gio object to simulate logind responses

        The core of the resulting object is the results_spec, which is a nested
        mapping. The outer layer maps session IDs to their own inner mappings. The
        inner layer maps string parameter names (e.g., "Leader") to arbitrary
        values. These are then returned as cached properties for each mocked
        Session.
        """

        mock_obj = Mock()

        mock_obj.bus_get_sync = Mock()
        mock_obj.BusType = Mock()
        mock_obj.BusType.SYSTEM = 'SYSTEM'
        mock_obj.DBusProxyFlags = Mock()
        mock_obj.DBusCallFlags = Mock()

        mock_obj.DBusProxy = Mock()
        mock_obj.DBusProxy.new_sync = Mock(
                side_effect=self._dbus_proxy_new_sync
        )

        # We are mocking this object -- nothing is protected from us!
        # pylint: disable=protected-access
        mock_obj._manager = Mock()
        mock_obj._manager.call_sync = Mock(
                side_effect=self._manager_call_sync
        )

        self._sessions: Mapping[str, Mock] = {}
        for session_id, spec in results_spec.items():
            self._sessions[session_id] = Mock()
            self._sessions[session_id].get_cached_property = Mock(
                    side_effect=partial(
                        self._session_get_cached_property,
                        spec
                    )
            )

        return mock_obj

    def _dbus_proxy_new_sync(self, *args):
        """Return either a mocked Manager or a mocked Session"""

        # We are mocking this object -- nothing is protected from us!
        # pylint: disable=protected-access
        if args[5] == "org.freedesktop.login1.Manager":
            return self._mocked_gio._manager
        if args[5] == "org.freedesktop.login1.Session":
            session_id_match = re.match(_SESSION_NODE_RE, args[4])
            if session_id_match is None:
                raise ValueError(f'invalid session node {args[4]}')
            return self._sessions[session_id_match.group(1)]
        raise KeyError('Unknown D-Bus interface: {args[5]}')

    def _manager_call_sync(self, *_):
        """Return a list of Session results with the set of spec'd IDs"""

        packed_result = Mock()
        packed_result.unpack = Mock(side_effect=lambda: [
            list(map(lambda x: [x], self._sessions.keys()))
        ])
        return packed_result

    def _session_get_cached_property(self, spec: Mapping[str, Any], *args):
        """Return a session API object that can return specified values"""

        packed_result = Mock()
        packed_result.get_string = Mock(
                side_effect=lambda: str(spec[args[0]])
        )
        packed_result.get_uint32 = Mock(
                side_effect=lambda: int(spec[args[0]])
        )

        # This is a special datatype for User, listed in the docs as (uo)
        # https://www.freedesktop.org/software/systemd/man/latest/org.freedesktop.login1.html
        if args[0] == "User":
            packed_result.unpack = Mock(side_effect=lambda: [
                    int(spec[args[0]])
            ])

        return packed_result

    def _assert_fully_exercised(self):
        """Run assertions to ensure that the entire GioMock was exercised"""

        self._mocked_gio.bus_get_sync.assert_called()

        self._mocked_gio.DBusProxy.new_sync.assert_any_call(
                ANY,
                ANY,
                ANY,
                'org.freedesktop.login1',
                '/org/freedesktop/login1',
                'org.freedesktop.login1.Manager',
                ANY
        )

        self._mocked_gio.DBusProxy.new_sync.assert_any_call(
                ANY,
                ANY,
                ANY,
                'org.freedesktop.login1',
                ANY,
                'org.freedesktop.login1.Session',
                ANY
        )

        # We are mocking this object -- nothing is protected from us!
        # pylint: disable=protected-access
        self._mocked_gio._manager.call_sync.assert_called_with(
                'ListSessions',
                ANY,
                ANY,
                ANY,
                ANY
        )

        for session_id, session in self._sessions.items():
            try:
                session.get_cached_property.assert_called()
            except AssertionError as e:
                raise AssertionError(f'Session ID {session_id}') from e

    #
    # Internal attributes used by test cases -- subclasses shouldn't use these
    #

    # The mocked Gio object created by _mock_gio()
    _mocked_gio: Mock

    # Cached list of sessions, to allow for later assertions
    _sessions: Mapping[str, Mock]


def load_tests(*_):
    """Implementation of the load_tests protocol

    https://docs.python.org/3/library/unittest.html#load-tests-protocol

    All of the test cases should be added by the test_scenario*.py files. No
    unit tests should be run directly from this common file.

    We ignore the 1st argument (loader), 2nd argument (standard_tests), and
    3rd argument (pattern) and substitute a totally custom (empty) TestSuite.
    """
    return TestSuite()



"""Common process table / cgroup testing logic shared across all scenarios"""


from typing import List, Mapping
from unittest import TestCase, TestSuite
from unittest.mock import Mock, mock_open, patch

import stop_idle_sessions.ps


class CgroupPidsTestCase(TestCase):
    """Unit testing for the ps module

    This TestCase is meant to be subclassed, NOT run directly. The load_tests
    function at the bottom of this module prevents it from being
    auto-discovered.
    """

    #
    # Subclasses need to override these methods
    #

    def _mock_process_specs(self) -> Mapping[int, str]:
        """Subclasses should override this method"""
        raise NotImplementedError('_mock_process_specs')

    def _expected_process_objects(self) -> List[stop_idle_sessions.ps.Process]:
        """Subclasses should override this method"""
        raise NotImplementedError('_expected_logind_sessions')

    #
    # Here are the actual test case methods -- these aren't usually overridden
    #

    def setUp(self):
        self._mock_psutil_process(self._mock_process_specs())

        mock_psutil_process_patcher = patch('psutil.Process',
                                            new=self._mocked_psutil_process)
        mock_psutil_process_patcher.start()
        self.addCleanup(mock_psutil_process_patcher.stop)

    def test_ps_interface_parsed_objects(self):
        """Ensure that processes are appropriately parsed from the cgroup"""

        expected_processes = self._expected_process_objects()
        actual_processes = list(
                stop_idle_sessions.ps.processes_in_scope_path(
                    "/user.slice/user-1000.slice/session-1024.scope",
                    open_func=self._mocked_open
                )
        )

        self._mocked_psutil_process.assert_called()
        self.assertListEqual(expected_processes, actual_processes)

    #
    # Internal methods used by test cases -- these should not be overridden
    #

    def _mock_psutil_process(self, process_specs: Mapping[int, str]):
        """Mock the sysfs directory file and process table for a set of PIDs

        Each PID may be associated with its own "command line" (e.g.
        /usr/sbin/sshd [...args...]).
        """

        self._mocked_processes = {}

        self._mocked_psutil_process = Mock(side_effect=self._psutil_process_init)
        for pid, cmdline in process_specs.items():
            self._mocked_processes[pid] = Mock()
            self._mocked_processes[pid].pid = pid
            self._mocked_processes[pid].cmdline = Mock(return_value=cmdline.split())
            self._mocked_processes[pid].environ = Mock(return_value={})

        self._mocked_open = mock_open(read_data=self._cgroup_procs_content)

    @property
    def _cgroup_procs_content(self) -> str:
        """File contents of the simulated cgroup.procs file

        The PIDs from this Mock are guaranteed to be sorted. This appears to
        be true for the real sysfs too (though it's not clear whether that is
        a specified behavior or an implementation detail).
        """
        return "".join(map(lambda x: f"{x}\n",
                           sorted(self._mocked_processes.keys())))

    def _psutil_process_init(self, *args, **_) -> Mock:
        """Mock the initialization of a Process by looking up a PID"""
        return self._mocked_processes[int(args[0])]

    #
    # Internal attributes used by test cases -- subclasses shouldn't use these
    #

    # The mocked psutil.Process object created by _mock_psutil_process()
    _mocked_psutil_process: Mock

    # Individually-mocked Process objects which have been cached
    _mocked_processes: Mapping[int, Mock]

    # A mocked variant of the builtin open() function
    _mocked_open: Mock


def load_tests(*_):
    """Implementation of the load_tests protocol

    https://docs.python.org/3/library/unittest.html#load-tests-protocol

    All of the test cases should be added by the test_scenario*.py files. No
    unit tests should be run directly from this common file.

    We ignore the 1st argument (loader), 2nd argument (standard_tests), and
    3rd argument (pattern) and substitute a totally custom (empty) TestSuite.
    """
    return TestSuite()



"""Main logic for the stop-idle-session loop"""


import argparse
import configparser
import datetime
from itertools import product
import logging
import logging.handlers
import re
import sys
import traceback
from typing import List, Mapping, NamedTuple, Optional, Tuple

from stop_idle_sessions.exception import SessionParseError
import stop_idle_sessions.getent
import stop_idle_sessions.logind
import stop_idle_sessions.ps
import stop_idle_sessions.ss
import stop_idle_sessions.tty
import stop_idle_sessions.x11


logger = logging.getLogger(__name__)
_DEFAULT_CONFIG_FILE = "/etc/stop-idle-sessions.conf"


class SessionProcess(NamedTuple):
    """Representation of a Process specifically inside of a Session"""

    # Generic Process details for this SessionProcess
    process: stop_idle_sessions.ps.Process

    # A (possibly empty) list of backend processes that this particular
    # process has tunneled back into
    tunneled_processes: List[stop_idle_sessions.ps.Process]

    # A (possibly empty) list of Sessions that contain the tunneled_processes
    # that this particular process has connected to
    tunneled_sessions: List['Session']

    def __eq__(self, other):
        if not hasattr(other, 'process'):
            return False
        if not hasattr(other.process, 'pid'):
            return False
        return self.process.pid == other.process.pid


class Session(NamedTuple):
    """Representation of an individual Session, combining various sources"""

    # Backend logind session object for this Session
    session: stop_idle_sessions.logind.Session

    # The TTY or PTY which is assigned to this session (or None)
    tty: Optional[stop_idle_sessions.tty.TTY]

    # The value of the DISPLAY environment variable for this particular
    # process, or None if none was assigned
    display: Optional[str]

    # The idletime (expressed as a timedelta) reported by the X11 Screen Saver
    # extension for this DISPLAY (or None if there was no DISPLAY)
    display_idle: Optional[datetime.timedelta]

    # The symbolic username corresponding to session.uid
    username: str

    # Collection of Process objects belonging to this Session
    processes: List[SessionProcess]

    def __eq__(self, other):
        if not hasattr(other, 'session'):
            return False
        if not hasattr(other.session, 'session_id'):
            return False
        return self.session.session_id == other.session.session_id

    def __str__(self):
        return self.string_representation()

    def string_representation(self,
                              idleness: Optional[datetime.timedelta] = None) -> str:
        """Provide a string representation of the Session, maybe with idleness"""

        tty_string: str = "notty"
        if self.tty is not None:
            tty_string = self.tty.name

        nr_processes: int = len(self.processes)
        nr_tunnels: int = sum(map(lambda p: len(p.tunneled_processes),
                                  self.processes))
        nr_sessions: int = sum(map(lambda p: len(p.tunneled_sessions),
                                   self.processes))

        backend_text: str = f"w/ {nr_processes} processes"
        if nr_tunnels > 0:
            backend_text += f" hosting {nr_tunnels}"
            if nr_sessions > 0:
                backend_text += f" to {nr_sessions} backend sessions"

        idleness_text: str = ""
        if idleness is not None:
            idleness_text = f" -- idle for {idleness.total_seconds() // 60} minutes"

        return (f'session id="{self.session.session_id}" '
                f'leader={self.session.leader} '
                f'user={self.username}@{tty_string} '
                f'({backend_text}){idleness_text}')


# Constructing the tree involves many local variables, necessarily
# pylint: disable-next=too-many-locals, too-many-branches
def load_sessions() -> List[Session]:
    """Construct an abstract Session/Process tree from system observations"""

    try:
        logind_sessions = stop_idle_sessions.logind.get_all_sessions()
        loopback_connections = stop_idle_sessions.ss.find_loopback_connections()
    except SessionParseError as err:
        logger.error('Problem while reading session and networking table '
                     'information: %s', err.message)
        raise err

    resolved_usernames: Mapping[int, str] = {}

    # Constructing the tree involves many layers of nesting, necessarily
    # pylint: disable=too-many-nested-blocks
    sessions: List[Session] = []
    display_collector = stop_idle_sessions.x11.X11DisplayCollector()
    for logind_session in logind_sessions:
        try:

            if logind_session.uid not in resolved_usernames:
                username = stop_idle_sessions.getent.uid_to_username(
                        logind_session.uid
                )
                resolved_usernames[logind_session.uid] = username

            session_processes: List[SessionProcess] = []
            ps_table = stop_idle_sessions.ps.processes_in_scope_path(
                    logind_session.scope_path
            )
            for process in ps_table:
                tunneled_processes: List[stop_idle_sessions.ps.Process] = []
                display_collector.add(logind_session.session_id, process)

                # Associate Processes thru loopback to other Processes
                for loopback_connection in loopback_connections:
                    client_processes = loopback_connection.client.processes
                    server_processes = loopback_connection.server.processes
                    for client_process in client_processes:
                        for server_process in server_processes:
                            if process == client_process:
                                if not server_process in tunneled_processes:
                                    tunneled_processes.append(server_process)

                session_processes.append(SessionProcess(
                        process=process,
                        tunneled_processes=tunneled_processes,
                        tunneled_sessions=[]
                ))

            session_tty: Optional[stop_idle_sessions.tty.TTY] = None
            # If a session is "lingering" (i.e., has no remaining Leader
            # process), then we cannot and do not trust its tty string.
            if logind_session.tty != "" and logind_session.leader != 0:
                session_tty = stop_idle_sessions.tty.TTY(
                        logind_session.tty
                )

            sessions.append(Session(
                    session=logind_session,
                    tty=session_tty,
                    display=None,
                    display_idle=None,
                    username=resolved_usernames[logind_session.uid],
                    processes=session_processes
            ))

        except SessionParseError as err:
            logger.warning('Could not successfully parse information related '
                           'to session id="%s": %s',
                           logind_session.session_id,
                           err.message)
            traceback.print_exc(file=sys.stderr)

    # Go back and resolve display information and idletimes for their Sessions
    sessions_prex11, sessions = sessions, []
    for session in sessions_prex11:
        try:
            display_info = display_collector.retrieve_least_display_idletime(
                    session.session.session_id
            )
            if display_info is None:
                sessions.append(session)
            else:
                sessions.append(Session(
                    session=session.session,
                    tty=session.tty,
                    display=display_info[0],
                    display_idle=display_info[1],
                    username=session.username,
                    processes=session.processes
                ))
        except SessionParseError as err:
            logger.warning('Could not successfully parse X11 info related '
                           'to session %s: %s',
                           session.string_representation(),
                           err.message)
            traceback.print_exc(file=sys.stderr)

    # Go back and resolve backend tunneled Processes to their Sessions
    for session_a, session_b in product(sessions, sessions):
        for process_a, process_b in product(session_a.processes,
                                            session_b.processes):
            for backend_process_a in process_a.tunneled_processes:
                if backend_process_a == process_b.process:
                    process_a.tunneled_sessions.append(session_b)

    return sessions


def skip_ineligible_session(session: Session,
                            excluded_users: Optional[List[str]] = None) -> Tuple[bool,
                                                                                 str]:
    """Check whether a session is ineligible for idleness timeout enforcement

    Returns True if this session meets any of the criteria for being
    "ineligible" (see README.md) and should not be processed further. If
    False, then the caller should continue processing.

    The returned string is a short descriptor for why the session was skipped.
    It is blank if the session was not skipped.
    """

    # Graphical sessions should be protected by screensavers, not idle
    # timeouts. (_Tunneled_ graphical sessions are a different story -- see
    # README.md for details.)
    # https://github.com/systemd/systemd/blob/v256.8/src/login/logind-session.c#L1650
    if session.session.session_type in ('x11', 'wayland', 'mir'):
        return True, 'graphical'

    # systemd-logind sessions without an assigned teletype (TTY/PTY) which
    # represent "noninteractive" sessions.
    if session.tty is None:
        return True, 'noninteractive'

    # systemd-logind sessions belonging to one of the "excluded_users" in the
    # provided list (if any).
    if excluded_users is not None and session.username in excluded_users:
        return True, f'excluded user ({session.username})'

    # systemd-logind session whose Scope Leader has been terminated. See
    # README.md for a discussion of why this is relevant.
    if session.session.leader == 0:
        return True, 'lingering'

    return False, ''


def compute_idleness_metric(session: Session,
                            now: datetime.datetime,
                            nested: bool = False) -> datetime.timedelta:
    """Determine the most "optimistic" idleness metric for the given Session

    By drawing from all of the potential sources of idleness/activity, find
    the one which gives the user the most opportunity to keep their session
    alive. For example, if the user's keystroke activity would indicate 10
    minutes of idleness, but their tunneled VNC session experienced activity
    in the past 5 minutes, then the overall idleness should only be reported
    as 5 minutes.

    Some data sources (e.g., a TTY's atime) provide an absolute timestamp
    instead of a timedelta. The caller is expected to supply the current
    timestamp (e.g., datetime.datetime.now() OR a fake value for testing) so
    that such timestamps may be converted into a timedelta.

    Idleness can involve the participation of a tunneled session. This
    analysis should never need to extend past "depth two" -- e.g., the
    idleness of an SSH session which has tunneled through to a VNC session may
    depend on the idleness of the VNC session. _However_, we never need to
    (and don't want to!) enter a recursive loop where a session relates to
    itself, or two sessions relate circularly. As a protection against this,
    the nested argument will keep this function from analyzing further Session
    entries after the first one.
    """

    minimum_idle: Optional[datetime.timedelta] = None
    determined_by: str = ""

    # The atime on a TTY for a terminal session is touched whenever the user
    # enters keyboard input.
    if (session.tty is not None and (minimum_idle is None or
                                     now - session.tty.atime < minimum_idle)):
        minimum_idle = now - session.tty.atime
        determined_by = f"atime on {session.tty.name}"

    # The mtime on a TTY for a terminal session is touched whenever the user
    # enters keyboard input (same as atime) *OR* whenever a program generates
    # standard output/error onto the screen.
    if (session.tty is not None and (minimum_idle is None or
                                     now - session.tty.mtime < minimum_idle)):
        minimum_idle = now - session.tty.mtime
        determined_by = f"mtime on {session.tty.name}"

    # The idleness of a session which contains running some running X11 server
    # (probably Xwayland or Xorg or Xvnc) can be influenced by the same
    # activity metric tracked by the X11 Screen Saver extension.
    if (session.display_idle is not None and (minimum_idle is None or
                                              session.display_idle < minimum_idle)):
        minimum_idle = session.display_idle
        determined_by = f"X11 idleness on DISPLAY={session.display}"

    # The idleness of a session which has tunneled into another session
    # (primarily VNC over SSH) can be influenced by the tunneled session.
    # However, as explained in the docstring, this must be guarded against
    # recursive loops by enforcing a maximum depth of 2 (i.e., the outer
    # session and the inner session).
    if not nested:
        for session_process in session.processes:
            for tunneled_session in session_process.tunneled_sessions:
                # Do NOT check the inner session for "eligibility"
                try:
                    inner_idle = compute_idleness_metric(tunneled_session,
                                                        now,
                                                        nested=True)
                    if minimum_idle is None or inner_idle < minimum_idle:
                        minimum_idle = inner_idle
                        determined_by = ("idleness of nested session " +
                                            tunneled_session.session.session_id)
                except SessionParseError:
                    # Just skip this attempt if it didn't work out
                    pass

    if minimum_idle is None:
        raise SessionParseError('Could not identify an idle duration for ' +
                                session.session.session_id)

    logger.debug("Computed idleness for %s (based on %s)",
                 session.string_representation(minimum_idle),
                 determined_by)

    return minimum_idle


# This is a bit complicated, but it's the glue for everything else.
# pylint: disable-next=too-many-branches, too-many-locals, too-many-statements
def main():
    """Overall main loop routine for this application"""
    parser = argparse.ArgumentParser(
            prog='stop_idle_sessions.main',
            description=("Stop idle `systemd-logind` sessions to prevent "
                         "interactive access from unattended terminals. "
                         "E.g., a laptop left unlocked in a coffee shop, "
                         "with an SSH session into an internal network "
                         "resource.")
    )
    parser.add_argument('-n', '--dry-run', action='store_true',
                        help=("Don't actually take any actions, but just log "
                              "(to stdout, not syslog) what would have "
                              "happened"))
    parser.add_argument('-s', '--syslog', action='store_true',
                        help=("Write logged messages (including debug logs, "
                              "if configured) to syslog instead of stdout. "))
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Increase verbosity to incorporate debug logs "
                             "(either to stdout during dry-run, or syslog "
                             "normally)")
    parser.add_argument('-c', '--config-file', action='store',
                        default=_DEFAULT_CONFIG_FILE,
                        help="Override the location of the configuration INI "
                             "format file")

    args = parser.parse_args()

    config = configparser.ConfigParser()
    config['stop-idle-sessions'] = {
            'dry-run': 'no',
            'syslog': 'no',
            'verbose': 'no',
            'debug-log': '',
            'excluded-users': '',
            'timeout': '15'
    }

    try:
        with open(args.config_file, "r", encoding='utf-8') as config_f:
            config.read_file(config_f, source=args.config_file)
    except OSError as err:
        # If it was the default file that failed to open, then just ignore the
        # failure. Otherwise, this is a fatal condition.
        if args.config_file != _DEFAULT_CONFIG_FILE:
            logger.error('Problem while reading a custom config file '
                         'located at %s: %s',
                         args.config_file,
                         str(err))
            raise err

    try:
        dry_run: bool = config.getboolean('stop-idle-sessions', 'dry-run')
        syslog: bool = config.getboolean('stop-idle-sessions', 'syslog')
        verbose: bool = config.getboolean('stop-idle-sessions', 'verbose')
        debug_log: str = config.get('stop-idle-sessions', 'debug-log')
        excluded_users: List[str] = list(map(lambda x: x.strip(),
                                             re.split(r'[,;:]',
                                                      config.get('stop-idle-sessions',
                                                                 'excluded-users'))))
        timeout: int = config.getint('stop-idle-sessions', 'timeout')
    except ValueError as err:
        logger.error('Problem while parsing arguments: %s',
                     str(err))
        raise err

    # Command-line argument can override the config bools
    if args.dry_run:
        dry_run = True
    if args.syslog:
        syslog = True
    if args.verbose:
        verbose = True

    if syslog:
        syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
        syslog_handler.ident = 'stop-idle-sessions '
        logger.addHandler(syslog_handler)

        if verbose and debug_log != "":
            logger.setLevel(logging.DEBUG)
            debug_handler = logging.FileHandler(debug_log, encoding='utf-8')
            debug_handler.setFormatter(
                    logging.Formatter(fmt='%(asctime)s %(message)s')
            )
            logger.addHandler(debug_handler)

    elif verbose:
        logging.basicConfig(level=logging.DEBUG)

    now = datetime.datetime.now()
    sessions = load_sessions()
    for session in sessions:
        skip_session, why = skip_ineligible_session(session, excluded_users)

        try:
            idletime = compute_idleness_metric(session, now)

            if idletime >= datetime.timedelta(seconds=60 * timeout):
                if skip_session:
                    logger.debug('Skipping %s session %s',
                                 why,
                                 session.string_representation(idletime))
                elif dry_run:
                    logger.warning('Not terminating (dry run) %s',
                                   session.string_representation(idletime))
                else:
                    logger.warning('Terminating %s',
                                   session.string_representation(idletime))
                    session.session.kill_session_leader()

        except SessionParseError as err:
            logger.debug('Could not determine idletime for %s (%s)',
                         session.string_representation(),
                         err.message)
            if not skip_session:
                raise err


if __name__ == "__main__":
    main()



"""X11 screen saver information to determine idle time"""


from collections import defaultdict
from datetime import timedelta
import os
import re
from typing import Dict, Optional, Set, Tuple

import Xlib.display
import Xlib.error

from .exception import SessionParseError
from .ps import Process


class X11DisplayCollector:
    """Collect related Process objects to determine X11 params across sessions

    There are often many Process objects associated with a SystemD scope or
    session. There may be one (or more!) instances of the DISPLAY or
    XAUTHORITY variables among them. Some of their commandlines may even
    provide clues as to these parameters.

    In some cases, such as for x11vnc, the full picture may not emerge until
    _multiple_ sessions have been processed. For this reason, the
    X11DisplayCollector is intended to provide a global repository that all
    session processes are fed into. It can then provide back an iteration of
    identified displays associated with each session.

    Once collected, these parameters can point to one or more DISPLAYs which
    may provide an idle time (via the X11 Screen Saver extension).
    """

    def __init__(self):
        # Each session identifier string may be associated with one or more
        # individual DISPLAY candidates.
        self._session_displays: Dict[str, Set[str]] = defaultdict(set)

        # Subsequently, each individual DISPLAY can be associated with one or
        # more individual XAUTHORITY candidates. Note that, for this
        # collection, the value set may (and generally should) include a None
        # value, to indicate that the DISPLAY should be attempted to be
        # accessed with no Xauthority (in addition to any other located
        # candidate values).
        self._display_xauthorities: Dict[str, Set[Optional[str]]] = defaultdict(set)

    def add(self, session: str, process: Process):
        """Add information from a Process and its session ID to tracking

        This will extract information from the given Process which will allow
        the "candidate tuples" list to expand and incorporate the new info.
        The Processes are not actually collected internally per se -- just
        relevant information.
        """

        # Try some specific command lines
        xserver_match = X11DisplayCollector.parse_xserver_cmdline(process.cmdline)
        x11vnc_match = X11DisplayCollector.parse_x11vnc_cmdline(process.cmdline)

        display: Optional[str] = None
        xauthority: Optional[str] = None

        if xserver_match[0] is not None:
            display = xserver_match[0]
        elif x11vnc_match is not None:
            display = x11vnc_match
        elif 'DISPLAY' in process.environ:
            display = process.environ['DISPLAY']

        if xserver_match[1] is not None:
            xauthority = xserver_match[1]
        elif 'XAUTHORITY' in process.environ:
            xauthority = process.environ['XAUTHORITY']

        if display is not None:
            self._session_displays[session].add(display)
            self._display_xauthorities[display].add(xauthority)
            if None not in self._display_xauthorities[display]:
                self._display_xauthorities[display].add(None)

    def retrieve_least_display_idletime(self, session: str) -> Optional[Tuple[str,
                                                                        timedelta]]:
        """Retrieve the smallest of DISPLAY idletimes, and the DISPLAY itself

        Why the smallest? We want to be as optimistic as possible about idle
        times to keep from terminating user processes without a good reason.
        Even if there is, say, a rogue process in a VNC session which is
        connected to some external place via X11 forwarding, we would rather
        that idletime be checked against both (perhaps surprisingly) than to
        incorrectly terminate a non-idle session.

        The first return value is the DISPLAY string, and the second is its
        idletime.
        """

        # Arbitrarily keep track of one (of possibly several)
        # SessionParseErrors, and raise it if no timedeltas are ever
        # successfully retrieved.
        any_exception: Optional[SessionParseError] = None
        result: Optional[Tuple[str, timedelta]] = None

        for display in self._session_displays[session]:
            for xauthority in self._display_xauthorities[display]:
                try:
                    candidate_idletime = X11DisplayCollector.retrieve_idle_time(
                            display,
                            xauthority
                    )
                    if candidate_idletime is not None:
                        if result is None:
                            result = (display, candidate_idletime)
                        elif candidate_idletime < result[1]:
                            result = (display, candidate_idletime)

                except SessionParseError as err:
                    # Given the choice: If an XAUTHORITY was determined, then
                    # trust the _new_ error. If no XAUTHORITY was determined, then
                    # keep the _old_ error (because it is very likely that a None
                    # XAUTHORITY would fail normally).
                    if any_exception is None or xauthority is not None:
                        any_exception = err

        if result is not None:
            return result
        if any_exception is not None:
            raise any_exception
        return None

    @staticmethod
    def parse_xserver_cmdline(cmdline: str) -> Tuple[Optional[str],
                                                     Optional[str]]:
        """Attempt to identify information from an X command line

        The first element of the returned tuple is a candidate DISPLAY, if one
        is found. The second is a candidate XAUTHORITY, if one is found.
        This works with Xvnc, Xwayland, and possibly others.
        """

        xserver_re = re.compile(r'^.*X[a-z]+\s+(:[0-9]+).*-auth\s+(\S+).*$')

        xserver_match = xserver_re.match(cmdline)
        if xserver_match is not None:
            return (xserver_match.group(1), xserver_match.group(2))

        return (None, None)

    @staticmethod
    def parse_x11vnc_cmdline(cmdline: str) -> Optional[str]:
        """Attempt to identify information from an x11vnc command line

        Similar to an X server, it may be possible to obtain information from
        an x11vnc command line. This is a little different because it can't
        usually provide XAUTHORITY information, but only XDISPLAY information.
        """

        x11vnc_re = re.compile(r'^.*x11vnc.*-display\s+(:[0-9]+).*$')

        x11vnc_match = x11vnc_re.match(cmdline)
        if x11vnc_match is not None:
            return x11vnc_match.group(1)

        return None

    @staticmethod
    def retrieve_idle_time(display: str,
                           xauthority: Optional[str] = None) -> Optional[timedelta]:
        """Retrieve the idle time (in milliseconds) for the given X11 DISPLAY"""

        # Crazy hack to try and work around this issue, reported by a _different
        # project_ (which has never made it into the python-xlib upstream):
        # https://github.com/asweigart/pyautogui/issues/202
        extensions = getattr(Xlib.display, 'ext').__extensions__
        if ('RANDR', 'randr') in extensions:
            extensions.remove(('RANDR', 'randr'))
        if ('XFIXES', 'xfixes') in extensions:
            extensions.remove(('XFIXES', 'xfixes'))

        try:
            if xauthority is not None:
                os.environ['XAUTHORITY'] = xauthority

            d = Xlib.display.Display(display)
            if d.has_extension('MIT-SCREEN-SAVER'):
                idle_time_ms = d.screen().root.screensaver_query_info().idle
                return timedelta(milliseconds=idle_time_ms)

            # The DISPLAY doesn't support the screen saver extension, which
            # means it is probably either forwarded (X11) or running a GDM
            # login session.
            return None

        except Xlib.error.DisplayConnectionError as err:
            raise SessionParseError(f'Could not connect to X11 display identified '
                                    f'by "{display}"') from err

        except Xlib.error.ConnectionClosedError as err:
            raise SessionParseError(f'Could not maintain a connection to the X11 '
                                    f'display identified by "{display}"') from err

        except AttributeError as err:
            raise SessionParseError(f'Cannot access attributes from X11 server '
                                    f'responses associated with display '
                                    f'"{display}", probably due to a broken or '
                                    f'erroneous connection') from err



"""Unit-test the set arithmetic function for unhashable List objects"""


from unittest import TestCase

from stop_idle_sessions import list_set


class CompareListSetsTestCase(TestCase):
    """Unit-test the set arithmetic function for unhashable List objects"""

    def test_identical_lists(self):
        """Two identical lists should be matched"""

        list_a = [1, 5, 8, 2, 4]
        list_b = [1, 5, 8, 2, 4]

        matchup = list_set.matchup_list_sets(list_a, list_b)
        similar = list_set.compare_list_sets(list_a, list_b)

        self.assertListEqual(sorted(matchup, key=lambda x: x[0]),
                             [(1, 1),
                              (2, 2),
                              (4, 4),
                              (5, 5),
                              (8, 8)])
        self.assertTrue(similar)

    def test_not_quite_identical_lists(self):
        """Two similar-length lists with a missing element"""

        list_a = [1, 5, 7, 2, 4]
        list_b = [1, 5, 8, 2, 4]

        matchup = list_set.matchup_list_sets(list_a, list_b)
        similar = list_set.compare_list_sets(list_a, list_b)

        self.assertListEqual(sorted(matchup, key=lambda x: x[0]),
                             [(1, 1),
                              (2, 2),
                              (4, 4),
                              (5, 5)])
        self.assertFalse(similar)

    def test_identical_but_misordered_lists(self):
        """Two identical but differently-ordered lists"""

        list_a = [5, 1, 2, 8, 4]
        list_b = [1, 5, 8, 2, 4]

        matchup = list_set.matchup_list_sets(list_a, list_b)
        similar = list_set.compare_list_sets(list_a, list_b)

        self.assertListEqual(sorted(matchup, key=lambda x: x[0]),
                             [(1, 1),
                              (2, 2),
                              (4, 4),
                              (5, 5),
                              (8, 8)])
        self.assertTrue(similar)

    def test_missing_a_element(self):
        """List A is missing an element"""

        list_a = [1, 5, 8, 2]
        list_b = [1, 5, 8, 2, 4]

        matchup = list_set.matchup_list_sets(list_a, list_b)
        similar = list_set.compare_list_sets(list_a, list_b)

        self.assertListEqual(sorted(matchup, key=lambda x: x[0]),
                             [(1, 1),
                              (2, 2),
                              (5, 5),
                              (8, 8)])
        self.assertFalse(similar)

    def test_missing_b_element(self):
        """List B is missing an element"""

        list_a = [1, 5, 8, 2, 4]
        list_b = [1, 5, 8, 2]

        matchup = list_set.matchup_list_sets(list_a, list_b)
        similar = list_set.compare_list_sets(list_a, list_b)

        self.assertListEqual(sorted(matchup, key=lambda x: x[0]),
                             [(1, 1),
                              (2, 2),
                              (5, 5),
                              (8, 8)])
        self.assertFalse(similar)



"""Set arithmetic on an unhashable List object"""


from itertools import product
from typing import Any, List, Tuple


def matchup_list_sets(list_a: List[Any],
                      list_b: List[Any]) -> List[Tuple[Any, Any]]:
    """Extract matches from two "sets" which are actually unhashable Lists

    A frequent pattern throughout this project -- both in the actual
    implementation and while testing it -- is to have two "sets" of elements
    which are actually just Lists. We want to compare the two "sets" and look
    for elements which are equivalent (according to __eq__). Normally this
    could be done with the set() type, but not for unhashable or mutable
    objects.

    Similar to the zip() function, this will return tuples of items from
    list_a and list_b. Not all elements will necessarily be represented --
    only those elements which have an equivalent match in the other list.

    This straightforward algorithm assumes that the lists aren't very big.
    """

    cartesian = product(list_a, list_b)
    matches = filter(lambda x: x[0] == x[1], cartesian)
    return list(matches)

def compare_list_sets(list_a: List[Any], list_b: List[Any]) -> bool:
    """Compare two "sets" which are actually just Lists of unhashable objects

    This is a further convenience around matchup_list_sets which simply
    returns True if the lists are equivalent (in the manner described for
    matchup_list_sets), or False otherwise.
    """

    if len(list_a) != len(list_b):
        return False
    return len(matchup_list_sets(list_a, list_b)) == len(list_a)



"""Common network loopback testing logic shared across all scenarios"""


from ipaddress import IPv4Address, IPv6Address
from os.path import basename
from typing import List, Set, Tuple, Union
from unittest import TestCase, TestSuite
from unittest.mock import Mock, patch

from stop_idle_sessions.list_set import compare_list_sets
import stop_idle_sessions.ps
import stop_idle_sessions.ss


class LoopbackConnectionTestCase(TestCase):
    """Unit testing for the ss module

    This TestCase is meant to be subclassed, NOT run directly. The load_tests
    function at the bottom of this module prevents it from being
    auto-discovered.
    """

    #
    # Subclasses need to override these methods
    #

    def _mock_raw_ss_output(self) -> str:
        """Subclasses should override this method"""
        raise NotImplementedError('_mock_raw_ss_output')

    def _expected_listening_ports(self) -> Set[int]:
        """Subclasses should override this method"""
        raise NotImplementedError('_expected_listening_ports')

    def _expected_peer_pairs(self) -> Set[Tuple[Union[IPv4Address,
                                                      IPv6Address], int]]:
        """Subclasses should override this method"""
        raise NotImplementedError('_expected_peer_pairs')

    def _expected_connections(self) -> List[stop_idle_sessions.ss.LoopbackConnection]:
        """Subclasses should override this method"""
        raise NotImplementedError('_expected_loopback_connections')

    #
    # Here are the actual test case methods -- these aren't usually overridden
    #

    def setUp(self):
        self._mocked_subprocess_run = Mock(
                side_effect=self._subprocess_run_with_check
        )

        mocked_subprocess_run_patcher = patch('subprocess.run',
                                              new=self._mocked_subprocess_run)
        mocked_subprocess_run_patcher.start()
        self.addCleanup(mocked_subprocess_run_patcher.stop)

    def test_three_stages_parsed_objects(self):
        """Ensure that each of the three parsing stages performs correctly

        This reaches a _bit_ more deeply into the code than a unit test
        perhaps should -- or at least it makes a few too many assertions --
        but it is useful to keep track of some of the inner workings to make
        sure everything is handled properly (without mocking internals).
        """

        invoke = stop_idle_sessions.ss.SSInvocation()
        invoke.run()

        expected_listening_ports = self._expected_listening_ports()
        actual_listening_ports=set(map(lambda s: s.port,
                                       invoke.listen_sockets))
        self.assertSetEqual(expected_listening_ports,
                            actual_listening_ports)

        expected_peer_pairs = self._expected_peer_pairs()
        actual_peer_pairs = set(map(lambda x: (x[1], x[2]),
                                    invoke.established_sockets))
        self.assertSetEqual(expected_peer_pairs,
                            actual_peer_pairs)

        expected_connections = self._expected_connections()
        actual_connections = invoke.loopback_connections
        self.assertTrue(compare_list_sets(expected_connections,
                                          actual_connections))

    #
    # Internal methods used by test cases -- these should not be overridden
    #

    def _subprocess_run_with_check(self, *args, **_) -> Mock:
        """Mock the subprocess.run call with a check for 'ss' command"""

        if basename(args[0][0]) != "ss":
            raise ValueError(f"Unexpected command {args[0]}")
        completed_process = Mock()
        completed_process.stdout = self._mock_raw_ss_output()
        return completed_process

    #
    # Internal attributes used by test cases -- subclasses shouldn't use these
    #

    # The mocked subprocess.run which will provide a particular stdout
    _mocked_subprocess_run: Mock


def load_tests(*_):
    """Implementation of the load_tests protocol

    https://docs.python.org/3/library/unittest.html#load-tests-protocol

    All of the test cases should be added by the test_scenario*.py files. No
    unit tests should be run directly from this common file.

    We ignore the 1st argument (loader), 2nd argument (standard_tests), and
    3rd argument (pattern) and substitute a totally custom (empty) TestSuite.
    """
    return TestSuite()



"""Unit testing for getent passwd command wrapper"""


from unittest import TestCase
from unittest.mock import Mock, patch

from stop_idle_sessions.exception import SessionParseError
import stop_idle_sessions.getent


class UidToUsernameTestCase(TestCase):
    """Unit-testing for the getent passwd command wrapper"""

    def test_successful_name_resolution(self):
        """Verify the module behavior when the user does exist"""
        completed_process = Mock()
        completed_process.stdout = "auser:x:1000:1000:A User:/home/auser:/bin/bash\n"
        completed_process.stderr = ""
        completed_process.returncode = 0

        with patch('subprocess.run',
                   new=Mock(return_value=completed_process)) as mock_run:
            actual_user = stop_idle_sessions.getent.uid_to_username(1000)
            self.assertEqual('auser', actual_user)
            mock_run.assert_called()

    def test_unsuccessful_name_resolution(self):
        """Verify the module behavior when the user does not exist"""
        completed_process = Mock()
        completed_process.stdout = ""
        completed_process.stderr = ""
        completed_process.returncode = 2

        with patch('subprocess.run',
                   new=Mock(return_value=completed_process)) as mock_run:
            with self.assertRaises(SessionParseError):
                stop_idle_sessions.getent.uid_to_username(1000)
            mock_run.assert_called()
